/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: config.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Support new config options
 *
 */

#include "defs.h"
#include <unistd.h>

#define EQUAL(s1, s2)           (strcmp((s1), (s2)) == 0)

char *configfilename = "/etc/pimd.conf";
/* candidate RP flags; global variables, initialized to 0's */
int CRPFlag = 0, CBSRFlag = 0;
struct g_rplist *confBSR = NULL;	/* the configured BSR */
u_char confBSRPrio = 0;		/* configured BSR priority */
struct g_rplist *CRP = NULL;	/* to hold the configured candidate RP address */
struct g_rplist BSR;		/* used to hold the bootstrap RP list */
u_int reg_rate_timer = REGISTER_RATE_TIMER;	/* initialize it */
u_int data_rate_timer = DATA_RATE_TIMER;	/* initialize it */
u_int reg_rate_limit = REGISTER_RATE_LIMIT;
u_int data_rate_limit = DATA_RATE_LIMIT;
u_int c_rp_adv_interval = C_RP_Adv_INTERVAL;	/* default initial setting */

static char * next_word(char **s);

void config_vifs_from_kernel()
{
	char ifbuf[BUFSIZ];
	struct ifconf ifc;
	struct ifreq *ifr;
	register struct vifentry *v;
	register vifi_t vifi;
	int i, nifs;
	__u32 addr, mask, subnet;
	short flags;
	int ifindex = 0;

	ifc.ifc_buf = ifbuf;
	ifc.ifc_len = sizeof(ifbuf);
	if (ioctl(igmp_socket, SIOCGIFCONF, (char *) &ifc) < 0)
		log(LOG_ERR, errno, "ioctl SIOCGIFCONF");
	nifs = ifc.ifc_len / sizeof(struct ifreq);

	for (i = nifs, ifr = ifc.ifc_req; i > 0; i--, ifr++) {
		struct vifaddr *vifa;
		addr = ((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr.s_addr;
		if (ifr->ifr_addr.sa_family != AF_INET)
			continue;

		if (ioctl(igmp_socket, SIOGIFINDEX, (char *) ifr) < 0) {
			log(LOG_ERR, errno, "ioctl SIOGIFINDEX for %s",
			    ifr->ifr_name);
			/* Not reached */
			continue;
		}
		ifindex = ifr->ifr_ifindex;
		if (ioctl(igmp_socket, SIOCGIFFLAGS, (char *) ifr) < 0) {
			log(LOG_ERR, errno, "ioctl SIOCGIFFLAGS for %s", ifr->ifr_name);
			/* Not reached */
			continue;
		}
		flags = ifr->ifr_flags;
		if (flags & IFF_POINTOPOINT) {
			if (ioctl(igmp_socket, SIOCGIFDSTADDR, (char *) ifr) < 0)
				log(LOG_ERR, errno, "ioctl SIOCGIFDSTADDR for %s", ifr->ifr_name);
			subnet = ((struct sockaddr_in *) &ifr->ifr_dstaddr)->sin_addr.s_addr;
			mask = 0xffffffff;
		} else {
			if ((flags & (IFF_LOOPBACK | IFF_MULTICAST)) != IFF_MULTICAST)
				continue;

			if (ioctl(igmp_socket, SIOCGIFNETMASK, (char *) ifr) < 0)
				log(LOG_ERR, errno, "ioctl SIOCGIFNETMASK for %s", ifr->ifr_name);
			mask = ((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr.s_addr;
			subnet = addr & mask;
		}
		if (!inet_valid_subnet(subnet, mask) || addr == subnet ||
		    addr == (subnet | ~mask) || addr == 0) {
			log(LOG_WARNING, 0, "ignoring %s, invalid address (%s) and/or mask (%08x)",
			    ifr->ifr_name, inet_fmt(addr, s1), ntohl(mask));
			continue;
		}
		for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
			struct vifaddr **vifap;
			int primary = 1;

			if (ifindex != v->ifindex)
				continue;
			for (vifap = &v->ifa; (vifa=*vifap) != NULL;
			     vifap = &vifa->next) {
				if (subnet == vifa->prefix &&
				    mask == vifa->mask) {
					if (addr == vifa->local)
						goto done;
					primary = 0;
				}
			}
			vifa = malloc(sizeof(*vifa));
			if (vifa == 0) {
				log(LOG_ERR, errno, "out of memory");
				continue;
			}

			vifa->next = NULL;
			vifa->local = addr;
			vifa->prefix = subnet;
			vifa->mask = mask;
			vifa->primary = primary;
			if (primary && ntohl(addr) > ntohl(v->prefsrc))
				v->prefsrc = addr;
			*vifap = vifa;
			break;
		}

done:
		if (vifi != numvifs)
			continue;

		if (numvifs == MAXVIFS) {
			log(LOG_WARNING, 0, "too many vifs, ignoring %s", ifr->ifr_name);
			continue;
		}
		v = &viflist[numvifs];
		v->ifindex = ifindex;

		vifa = malloc(sizeof(*vifa));
		if (vifa == 0) {
			log(LOG_ERR, errno, "out of memory");
			continue;
		}
		vifa->next = NULL;
		vifa->local = addr;
		vifa->prefix = subnet;
		vifa->mask = mask;
		vifa->primary = 1;
		v->ifa = vifa;
		v->prefsrc = addr;

		v->flags = 0;
		v->flags |= VIFF_NOMRT | VIFF_DR;
		v->metric = 1;
		v->threshold = MINTTL;

		strncpy(v->name, ifr->ifr_name, IFNAMSIZ);
		v->rate_limit = DEFAULT_PHY_RATE_LIMIT;

		numvifs++;
		if (!(flags & IFF_UP))
			v->flags |= VIFF_DOWN;

		log(LOG_INFO, 0,
		    "installing %s (%s on subnet %s) if:%d as vif #%d",
		    v->name, inet_fmt(addr, s1),
		    inet_fmts(subnet, mask, s2),
		    v->ifindex,
		    numvifs - 1);
	}
	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		if (v->ifa == NULL)
			v->flags |= VIFF_DOWN;
	}
}

#define UNKNOWN -1
#define EMPTY 1
#define PHYINT 2
#define CANDIDATE_RP 3
#define BOOTSTRAP_RP 4
#define REG_THRESHOLD 5
#define DATA_THRESHOLD 6
#define TUNNEL	7
#define RP_ADDRESS 8
#define MROUTE_STATIC  9
#define RP_DISC_FILTER  10

/* function name: wordToOption
   input: char *word, a pointer to the word
   output: int; a number corresponding to the code of the word
   operation: converts the result of the string comparisons into numerics.
   comments: called by config_vifs_from_file()
 */
static int wordToOption(char *word)
{
	if (EQUAL(word, ""))
		return EMPTY;
	if (EQUAL(word, "phyint"))
		return PHYINT;
	if (EQUAL(word, "candidate_rp"))
		return CANDIDATE_RP;
	if (EQUAL(word, "bootstrap_router"))
		return BOOTSTRAP_RP;
	if (EQUAL(word, "switch_register_threshold"))
		return REG_THRESHOLD;
	if (EQUAL(word, "switch_data_threshold"))
		return DATA_THRESHOLD;
	if (EQUAL(word, "rp_address"))
		return RP_ADDRESS;
	if (EQUAL(word, "rp_discovery_filter"))
		return RP_DISC_FILTER;
	if (EQUAL(word, "tunnel"))
		return TUNNEL;
	if (EQUAL(word, "mroute"))
		return MROUTE_STATIC;
	return UNKNOWN;
}

/* function name: parse_phyint
   input: char *s, pointing to the parsing point of the file
   output: int (TRUE if the parsing was successful, o.w. FALSE)
   operation: parses the physical interface file configurations, if any.
   comments: called by config_vifs_from_file()
 */
static int parse_phyint(char *s)
{
	char *w, c;
	vifi_t vifi;
	struct vifentry *v;
	u_int n;
	if (EQUAL((w = next_word(&s)), "")) {
		log(LOG_WARNING, 0, "missing phyint name in %s",
		    configfilename);
		return (FALSE);
	}			/* if empty */
	for (vifi = 0, v = viflist; (vifi < numvifs); ++vifi, ++v) {
		if (EQUAL(v->name, w))
			break;
	}
	if (vifi == numvifs) {
		log(LOG_WARNING, 0, "phyint %s in %s is not a configured interface",
		    v->name, configfilename);
		return (FALSE);
	}

	while (!EQUAL((w = next_word(&s)), "")) {
		if (EQUAL(w, "disable")) {
			v->flags |= VIFF_DISABLED;
			log(LOG_WARNING, 0, "Disabling %s (Vif: %d)",
			    v->name, vifi);
		} else if (EQUAL(w, "prefsrc")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing source address for phyint %s in %s",
				    v->name, configfilename);
			else {
				__u32 addr = inet_parse(w);
				if (addr != INADDR_NONE)
					v->prefsrc = addr;
			}
		} else if (EQUAL(w, "metric")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing metric for phyint %s in %s",
				    v->name, configfilename);
			else if (sscanf(w, "%u", &n) != 1 || n < 1)
				log(LOG_WARNING, 0, "invalid metric '%s' for phyint %s in %s",
				 w, v->name, configfilename);
			else
				v->metric = n;
		} else if (EQUAL(w, "threshold")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing threshold for phyint %s in %s",
				    v->name, configfilename);
			else if (sscanf(w, "%u%c", &n, &c) != 1 || n < 1 || n > 255)
				log(LOG_WARNING, 0, "invalid threshold '%s' for phyint %s in %s",
				 w, v->name, configfilename);
			else
				v->threshold = n;
		}
	}			/* if not empty */
	return (TRUE);
}

int parse_prefix(char *w, __u32 *addr, int *plen)
{
	char *ptr = strchr(w, '/');

	*plen = 32;
	if (ptr) {
		*ptr = 0;
		ptr++;
	}
	*addr = inet_parse(w);
	if (*addr == INADDR_NONE)
		return FALSE;
	if (ptr) {
		if (sscanf(ptr, "%u", plen) != 1)
			return FALSE;
		if (*plen > 32)
			return FALSE;
	}
	return TRUE;
}


/* function name: parse_rp_address
   input: char *s
   output: int (TRUE if the parsing was successful, o.w. FALSE)
   operation: parses the RP information for PIMv1.
   The general form is: 'rp address <ip-addr>'.
   comments: called by config_vifs_from_file()
 */
int parse_rp_address(char *s)
{
	char *w;
	struct g_rplist *list;
	__u32 addr;
	__u32 group;
	int glen;
	int ann = 0;
	int dense = 0;
	int version = 0;
	int scope = -1;
	int override = 0;

	if (EQUAL((w = next_word(&s)), ""))
		return FALSE;
	if (parse_prefix(w, &group, &glen) == FALSE)
		return FALSE;
	if (glen < 4)
		return FALSE;
	if (EQUAL((w = next_word(&s)), ""))
		return FALSE;
	if (!EQUAL(w, "dense")) {
		addr = inet_parse(w);
		if (addr == INADDR_NONE)
			return FALSE;
		if (EQUAL((w = next_word(&s)), "announce")) {
			ann = 1;
			if (EQUAL((w = next_word(&s)), ""))
				return FALSE;
			if (sscanf(w, "%d", &scope) != 1 || scope > 255)
				return FALSE;
			w = next_word(&s);
		}
	} else {
		dense = 1;
		addr = 0;
		ann = 0;
		scope = 0;
		w = next_word(&s);
	}
	if (EQUAL(w, "override")) {
		override = 1;
		w = next_word(&s);
	}
	if (EQUAL(w, "version")) {
		if (!EQUAL((w = next_word(&s)), ""))
			return FALSE;
		if (sscanf(w, "%d", &version) != 1)
			return FALSE;
		w = next_word(&s);
	}
	if (!EQUAL(w, ""))
		return FALSE;

	list = (struct g_rplist *) alloc_mem(sizeof(struct g_rplist));
	list->rpaddr = addr;
	list->group = group;
	list->grpmasklen = glen;
	list->grpmask = htonl((0xFFFFFFFF<<(32-glen)) & 0xFFFFFFFF);
	list->timer = 0;
	list->expires = 0xFFFFFFFF;
	list->info_source = 0;
	if (!version) {
		if (ann)
			version = 3;
		else
			version = 1;
	}
	list->pimversion = version;
	list->dense = dense;
	list->announce = ann;
	list->scope = scope;
	list->override = override;
	list->next = auto_rp_set;
	auto_rp_set = list;
	dirty_bits |= 4;
	return TRUE;
}

int parse_rp_discovery_filter(char *s)
{
	char *w;
	struct rp_afilter *a, **l;
	__u32 addr = 0;
	__u32 group;
	__u32 origin = 0;
	int glen;
	int alen = 0;
	int olen = 0;
	int flag = 0;

	if (EQUAL((w = next_word(&s)), ""))
		return FALSE;
	if (EQUAL(w, "any")) {
		glen = 4;
		group = htonl(0xE0000000);
	} else {
		if (parse_prefix(w, &group, &glen) == FALSE)
			return FALSE;
	}
	if (glen < 4)
		return FALSE;
	if (EQUAL((w = next_word(&s)), "rp")) {
		if (EQUAL((w = next_word(&s)), ""))
			return FALSE;
		if (parse_prefix(w, &addr, &alen) == FALSE)
			return FALSE;
		w = next_word(&s);
	}
	if (EQUAL(w, "agent")) {
		if (EQUAL((w = next_word(&s)), ""))
			return FALSE;
		if (parse_prefix(w, &origin, &olen) == FALSE)
			return FALSE;
		w = next_word(&s);
	}
	if (EQUAL(w, "deny")) {
		flag = 1;
	} else if (EQUAL(w, "permit")) {
		flag = 0;
	} else if (!EQUAL(w, ""))
		return FALSE;

	a = (struct rp_afilter *) alloc_mem(sizeof(struct rp_afilter));
	a->rpaddr = addr;
	if (alen)
		a->rpmask = htonl((0xFFFFFFFF<<(32-alen)) & 0xFFFFFFFF);
	else
		a->rpmask = 0;
	a->group = group;
	a->grplen = glen;
	a->grpmask = htonl((0xFFFFFFFF<<(32-glen)) & 0xFFFFFFFF);
	a->oaddr = origin;
	if (olen)
		a->omask = htonl((0xFFFFFFFF<<(32-olen)) & 0xFFFFFFFF);
	else
		a->omask = 0;
	a->flag = flag;
	a->next = NULL;
	
	for (l=&rp_discovery_filter; *l; l = &(*l)->next)
		/* NOTHING */;
	*l = a;
	dirty_bits |= 4;
	return TRUE;
}

#if 0
int parse_mroute(char *s)
{
	char *w;
	struct mrt_static *a, **l;
	__u32 addr;
	int alen;
	int flag = 0;
	int iface = -1;

	if (EQUAL((w = next_word(&s)), ""))
		return FALSE;
	if (parse_prefix(w, &addr, &alen) == FALSE)
		return FALSE;
	if (!EQUAL((w = next_word(&s)), "")) {
	}

	a = (struct mrt_static *) alloc_mem(sizeof(struct mrt_static));
	a->addr = addr;
	if (alen)
		a->mask = htonl((0xFFFFFFFF<<(32-alen)) & 0xFFFFFFFF);
	else
		a->mask = 0;
	a->next = NULL;
	
	for (l=&mrt_static_routes; *l; l = &(*l)->next)
		/* NOTHING */;
	*l = a;
	dirty_bits |= 4;
	return TRUE;
}
#endif

#ifdef NO_ANK_FIX

/* function name: parse_tunnel
   input: char *s
   output: int (TRUE if the parsing was successful, o.w. FALSE)
   operation: parses the tunnel information.
   The general form is: 'tunnel <local-addr> <remote-addr> metric <nn> threshold <nn> rate_limit <nn>'.
   comments: called by config_vifs_from_file()
 */
int parse_tunnel(char *s)
{
	char *w, *name;
	/*
	 * To fill de vifentry
	 */
	unsigned long local, remote;
	unsigned int metric = 1;
	unsigned char threshold = 0;
	unsigned int rate_limit = 0;

	struct g_rplist *list, *prev;
	struct vifentry *v;
	unsigned long addr;


	w = next_word(&s);
	if ((local = inet_parse(w)) == 0xffffffff) {
		log(LOG_WARNING, 0, "%s: Not a valid IPv4 internet address\n", w);
		return (FALSE);
	}
	/*
	   if ( local_address(local)==MAXVIFS ) {
	   log(LOG_WARNING,0,"%s: Is not a local IPv4 address\n", w );
	   return(FALSE);
	   }
	 */


	w = next_word(&s);
	if ((remote = inet_parse(w)) == 0xffffffff) {
		log(LOG_WARNING, 0, "%s: Not a valid IPv4 internet address\n", w);
		return (FALSE);
	}
	while (!EQUAL((w = next_word(&s)), "")) {
		if (EQUAL(w, "metric")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing metric value \n");
			else if (sscanf(w, "%u", &metric) != 1)
				log(LOG_WARNING, 0, "invalid metric %s", w);
			else {
				log(LOG_DEBUG, 0, "tunnel metric is %u", metric);
			}
		}
		/* if metric */ 
		else if (EQUAL(w, "threshold")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing threshold value \n");
			else if (sscanf(w, "%u", &threshold) != 1)
				log(LOG_WARNING, 0, "invalid threshold %s", w);
			else {
				log(LOG_DEBUG, 0, "tunnel threshold is %u", threshold);
			}
		}
		/* if threshold */ 
		else if (EQUAL(w, "rate_limit")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing rate_limit value \n");
			else if (sscanf(w, "%u", &rate_limit) != 1)
				log(LOG_WARNING, 0, "invalid rate_limit %s", w);
			else {
				log(LOG_DEBUG, 0, "tunnel rate_limit is %u", metric);
			}
		}
		/* if rate_limit */ 
		else if (EQUAL(w, "name")) {
			if (EQUAL((name = next_word(&s)), "")) {
				log(LOG_WARNING, 0, "missing name value \n");
				return (FALSE);
			}
			log(LOG_DEBUG, 0, "tunnel name is %s", name);
		}
		/* if rate_limit */ 
		else {
			log(LOG_WARNING, 0, "Invalid parameter: %s", w);
			return (FALSE);
		}
	}
	if (numvifs == MAXVIFS) {
		log(LOG_WARNING, 0, "Too many vifs already");
		return (FALSE);
	}
	v = &viflist[numvifs];
	v->flags = VIFF_TUNNEL | VIFF_NOMRT;
	if (local > remote)
		v->flags |= VIFF_DR;
	v->threshold = threshold;
	v->local = local;
	v->subnet = 0;
	v->subnetmask = 0;
	v->broadcast = 0;
	strcpy(v->name, name);
	v->rate_limit = rate_limit;
	v->remote = remote;	/* for tunnels */
	numvifs++;
	return (TRUE);
}

#endif

/* function name: parse_candidateRP
   input: char *s
   output: int (TRUE if the parsing was successful, o.w. FALSE)
   operation: parses the candidate RP information.
   The general form is: 'candidate_rp <local-addr> [time <number>]'.
   comments: called by config_vifs_from_file()
 */
static int parse_candidateRP(char *s)
{
	u_int time;
	char *w;
	u_long local = 0;

	if (debug > 3)
		log(LOG_DEBUG, 0, "CANDIDATE RP\n");

	CRPFlag = 1;		/* this is a global flag to trigger periodic
				 * candidate RPs; should be initialized to 0.
				 */
	/* may create a linked list for all groups of candidacy
	 * may use a group and a group mask too.
	 * Currently, candidacy is considered for all groups.
	 */

	while (!EQUAL((w = next_word(&s)), "")) {
		if (EQUAL(w, "time")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing c_rp_adv_interval time\n");
			else if (sscanf(w, "%u", &time) != 1)
				log(LOG_WARNING, 0, "invalid time %s", w);
			else {
				c_rp_adv_interval = time;
				log(LOG_DEBUG, 0, "c_rp_adv_interval is %u", time);
			}
		}
		/* if time */ 
		else if (!local) {
			if ((local = inet_parse(w)) == 0xffffffff ||
			    !inet_valid_host(local))
				/* invalid add */
				log(LOG_WARNING, 0, "C-RP Add not provided '%s' in %s",
				    w, configfilename);
			if (local_address(local) == MAXVIFS) {
				log(LOG_WARNING, 0, "C-RP Add non local '%s' in %s",
				    w, configfilename);

				CRPFlag = 0;	/* misconfiguration, so reset */
				return (FALSE);
			}
		}
		/* if ! local */ 
		else {
			log(LOG_WARNING, 0, "invalid parameter\n");
			return (FALSE);
		}
	}			/* while not empty */

	CRP = (struct g_rplist *) alloc_mem(sizeof(struct rplist));
	/* if the address is not provided, default to the max local */
	CRP->rpaddr = local ? local : max_local_address();

	if (debug > 3)
		log(LOG_DEBUG, 0, "Candidate RP is %s",
		    inet_fmt(CRP->rpaddr, s1));
	return (TRUE);
}

/* function name: parseBSR
   input: char *s
   output: int
   operation: parse BSR configured information.
   General form: 'bootstrap_router <local-addr> [priority <number>]'.
   comments: called by config_vifs_from_file()
 */
static int parseBSR(char *s)
{
	char *w;
	u_long local = 0;
	u_int priority;

	if (debug > 3)
		log(LOG_DEBUG, 0, "Candidate BOOTSTRAP ROUTER\n");

	CBSRFlag = 1;		/* global flag used to trigger bootrap
				   candidate advs in the queries; 
				   should be initialized to 0 */

	while (!EQUAL((w = next_word(&s)), "")) {
		if (EQUAL(w, "priority")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "\nMissing priority; default is 0\n");
			else if (sscanf(w, "%u", &priority) != 1)
				log(LOG_WARNING, 0, "invalid priority %s; defVal is 0", w);
			else {
				confBSRPrio = (u_char) priority;
				if (debug > 3)
					log(LOG_DEBUG, 0, "BSR Priority is %x", confBSRPrio);
			}
		}
		/* if priority */ 
		else if (!local) {
			if ((local = inet_parse(w)) == 0xffffffff ||
			    !inet_valid_host(local))
				/* invalid add */
				log(LOG_WARNING, 0, "BSR Address not provided '%s' in %s",
				    w, configfilename);
			if (local_address(local) == MAXVIFS) {
				log(LOG_WARNING, 0, "BSR Add non local '%s' in %s",
				    w, configfilename);

				CBSRFlag = 0;	/* misconfiguration, so reset */
				return (FALSE);
			}
		}
		/* if ! local */ 
		else {
			log(LOG_WARNING, 0, "invalid parameter\n");
			return (FALSE);
		}
	}			/* while not empty */

	confBSR = (struct g_rplist *) alloc_mem(sizeof(struct rplist));
	/* if the address is not provided, default to the max local */
	confBSR->rpaddr = local ? local : max_local_address();

	if (debug > 3)
		log(LOG_DEBUG, 0, "Candidate Bootstrap Router is %s",
		    inet_fmt(confBSR->rpaddr, s1));
	return (TRUE);
}

/* function name: parse_reg_threshold
   input: char *s
   output: int (TRUE if successful, FALSE o.w.)
   operation: reads and assigns the switch to the spt threshold
   due to registers for the router, if used as RP.
   Maybe extended to support different thresholds
   for different groups(prefixes).
   General form: 
   'switch_register_threshold [count <number> time <number>]'.
   comments: called by config_vifs_from_file()
 */
static int parse_reg_threshold(char *s)
{
	char *w;
	u_int count;
	u_int time;

	while (!EQUAL((w = next_word(&s)), "")) {
		if (EQUAL(w, "count")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing reg_rate count\n");
			else if (sscanf(w, "%u", &count) != 1)
				log(LOG_WARNING, 0, "invalid count %s", w);
			else {
				reg_rate_limit = count;
				log(LOG_DEBUG, 0, "reg_rate_limit is %u", count);
			}
		}
		/* if count */ 
		else if (EQUAL(w, "time")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing reg_rate time\n");
			else if (sscanf(w, "%u", &time) != 1)
				log(LOG_WARNING, 0, "invalid time %s", w);
			else {
				reg_rate_timer = time;
				log(LOG_DEBUG, 0, "reg_rate_time is %u", time);
			}
		}
		/* if time */ 
		else {
			log(LOG_WARNING, 0, "invalid parameter\n");
			return (FALSE);
		}
	}			/* while not empty */
	return (TRUE);
}

/* function name: parse_data_threshold
   input: char *s
   output: int
   operation: reads and assigns the switch to the spt threshold
   due to data packets, if used as DR.
   General form: 
   'switch_data_threshold [count <number> time <number>]'.
   comments: called by config_vifs_from_file()
 */
static int parse_data_threshold(char *s)
{
	char *w;
	u_int count;
	u_int time;

	while (!EQUAL((w = next_word(&s)), "")) {
		if (EQUAL(w, "count")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing data_rate count\n");
			else if (sscanf(w, "%u", &count) != 1)
				log(LOG_WARNING, 0, "invalid count %s", w);
			else {
				data_rate_limit = count;
				log(LOG_DEBUG, 0, "reg_data_limit is %u", count);
			}
		}
		/* if count */ 
		else if (EQUAL(w, "time")) {
			if (EQUAL((w = next_word(&s)), ""))
				log(LOG_WARNING, 0, "missing data_rate time\n");
			else if (sscanf(w, "%u", &time) != 1)
				log(LOG_WARNING, 0, "invalid time %s", w);
			else {
				data_rate_timer = time;
				log(LOG_DEBUG, 0, "data_rate_time is %u", time);
			}
		}
		/* if time */ 
		else {
			log(LOG_WARNING, 0, "invalid parameter\n");
			return (FALSE);
		}
	}			/* while not empty */
	return (TRUE);
}

void config_vifs_from_file()
{
	FILE *f;
	char linebuf[100];
	char *w, *s;
	struct ifconf ifc;
	int nifs, option;

	if ((f = fopen(configfilename, "r")) == NULL) {
		if (errno != ENOENT)
			log(LOG_WARNING, errno, "can't open %s",
			    configfilename);
		return;
	}
	ifc.ifc_buf = NULL;
	ifc.ifc_len = BUFSIZE;
	if (ioctl(igmp_socket, SIOCGIFCONF, (char *) &ifc) < 0)
		log(LOG_ERR, errno, "ioctl SIOCGIFCONF");
	nifs = ifc.ifc_len / sizeof(struct ifreq);

	while (fgets(linebuf, sizeof(linebuf), f) != NULL) {
		s = linebuf;
		w = next_word(&s);
		option = wordToOption(w);
		switch (option) {
		case EMPTY:
			continue;
			break;
		case PHYINT:
			parse_phyint(s);
			break;
		case CANDIDATE_RP:
			parse_candidateRP(s);
			break;
		case BOOTSTRAP_RP:
			parseBSR(s);
			break;
		case REG_THRESHOLD:
			parse_reg_threshold(s);
			break;
		case DATA_THRESHOLD:
			parse_data_threshold(s);
			break;
#ifdef NO_ANK_FIX
		case TUNNEL:
			parse_tunnel(s);
			break;
#endif
#if 0
		case MROUTE_STATIC:
			parse_mroute(s);
			break;
#endif
		case RP_ADDRESS:
			parse_rp_address(s);
			break;
		case RP_DISC_FILTER:
			parse_rp_discovery_filter(s);
			break;
		default:
			log(LOG_WARNING, 0, "unknown command '%s' in %s",
			    w, configfilename);
		}
	}

	if (auto_rp_set) {
		struct g_rplist *r, **rp;
		extern int autorp_announce;

		for (rp=&auto_rp_set; (r=*rp) != NULL; ) {
			if (r->announce) {
				if (local_address(r->rpaddr) == MAXVIFS) {
					*rp = r->next;
					log(LOG_INFO, 0, "announced rp %s is not me; deleting\n",
					    inet_fmt(r->rpaddr, s1));
					free(r);
					continue;
				} else {
					autorp_announce++;
				}
			}
			rp = &r->next;
		}
	}

	fclose(f);
}


static char * next_word(char **s)
{
	char *w;

	w = *s;
	while (*w == ' ' || *w == '\t')
		w++;

	*s = w;
	for (;;) {
		switch (**s) {
		case ' ':;
		case '\t':
			**s = '\0';
			(*s)++;
			return (w);
		case '\n':;
		case '#':
			**s = '\0';
			return (w);
		case '\0':
			return (w);
		default:
			if (isascii(**s) && isupper(**s))
				**s = tolower(**s);
			(*s)++;
		}
	}
}
