/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: defs.h
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */

/* Source code compatability */
#ifdef __linux__
#define ip_sum	ip_csum
#endif


#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <memory.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifndef __linux__
#include "../kernel/net/soioctl.h"
#endif
#ifdef __linux__
#include <sys/uio.h>
#endif
#include <sys/ioctl.h>
#include <linux/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#ifndef __linux__
#include <netinet/ip_var.h>
#endif
#include <sys/syscall.h>
#ifndef __linux__
#include "../kernel/netinet/igmp.h"
#include "../kernel/netinet/ip_mroute.h"
#else
#include <netinet/igmp.h>
#include <linux/mroute.h>
#endif
#include <time.h>

#include "pim.h"

#ifndef __P
#ifdef __STDC__
#define __P(x)  x
#else
#define __P(x)  ()
#endif
#endif

typedef void (*cfunc_t) __P((void *));
typedef void (*ihfunc_t) __P((int, fd_set *));


/* main.c */
extern u_long wall_time;
extern int debug;
void log(int severity, int syserr, char *format, ...);
extern void * alloc_mem(int size);

/* kern.c */
extern u_int numberOfCaches;
extern u_int numberOfAddMfcs;
void k_joinleave(int so, struct vifentry *v, __u32 mc, int joinleave);
void k_igmp_hdr_include(int);
void k_igmp_set_rcvbuf(int);
void k_igmp_set_loop(u_char b);
extern void k_pim_hdr_include(int);
extern void k_pim_set_rcvbuf(int);
extern void k_pim_set_ttl(u_char);
extern void k_pim_set_loop(u_char);
extern void k_pim_set_if(__u32);
void k_pim_set_rxinfo(u_char t);
void k_igmp_set_rxinfo(u_char t);
extern void k_init_pim(void);
extern void k_stop_pim(void);
void k_add_vif(vifi_t);
void k_del_vif(vifi_t);
void k_chg_viff(vifi_t);
extern int k_del_mrt(struct mrtentry *);
extern int k_chg_mrtf(struct mrtentry *);
extern u_long k_get_sg(struct mrtentry *);

/* routesock.c */
int k_req_incoming(struct rpfctl *);

/* config.c */
extern char *configfilename;
extern int CRPFlag;		/* candidate RP flag */
extern int CBSRFlag;		/* candidate BSR flag */
extern u_int reg_rate_limit;
extern u_int data_rate_limit;
extern u_int reg_rate_timer;
extern u_int data_rate_timer;
extern u_int c_rp_adv_interval;
extern struct g_rplist BSR;
extern struct g_rplist *CRP;
extern struct g_rplist *confBSR;	/* to hold the configured BSR address */
extern u_char confBSRPrio;	/* configured BSR priority */
extern void config_vifs_from_kernel(void);
extern void config_vifs_from_file(void);

/* igmp.c */
extern char pimv1_send_buf[2 * MAX_IP_PACKET_LEN];
extern char igmp_send_buf[128];
extern char dvmrp_send_buf[MAX_IP_PACKET_LEN];
extern int igmp_socket;
extern u_long allhosts_group;
extern u_long allrouters_group;
extern void init_igmp(void);
void igmp_read(int f, fd_set *rfd);
extern void accept_igmp(int);
extern void send_igmp(char *buf,
	       __u32 src,
	       __u32 dst,
	       int type,
	       int code,
	       int datalen,
	       int ifindex,
	       int mc_loop,
	       __u8 ttl);
extern void accept_mtrace_request(u_long src,
				  u_long dst,
				  u_long group,
				  char *data,
				  u_int no,
				  int datalen);
extern void accept_mtrace_response(char *buf, int len);
extern void accept_leave_message(u_long src, u_long dst);
extern int ip_encap(char *buf, unsigned long src, unsigned long dst, int datalen);

/* pim.c */
extern int pim_ifindex;
extern __u8 pim_recvttl;
extern char pim_recv_buf[MAX_IP_PACKET_LEN];
extern char pim_send_buf[MAX_IP_PACKET_LEN];
extern int pim_socket;
extern __u32 allpimrouters_group;
extern void init_pim(void);
void send_pim(char *buf, __u32 src, __u32 dst, int type, int datalen, int ifindex, __u8 ttl);
extern void send_pim_v1(char *buf, __u32 src, __u32 dst, int type, int datalen,
			int ifindex, int mc_loop, __u8 ttl);

/* bootstrap.c */
extern __u32 hashMask;

/* route.c */
extern u_char LclBSRPrio;
extern u_char AxptAnyRPSet;
extern u_char ElectedBSR;
extern u_int RPSetTimer;
extern u_char HashMaskLen;
extern u_short oifTimeout[MAXVIFS];
extern struct srcentry *srchash[256];
extern struct grpentry *grphash[256];
extern struct mrtentry *mrthash[256];
extern u_char jpholdtime[];
extern u_short c_rp_holdtime;
extern u_int C_RP_Adv_timer;
extern u_int C_RP_Timeout;
extern struct g_rplist *auto_rp_set;
extern struct rplist *list, *prev, *listHead;
void add_leaf(vifi_t vifi, __u32 group);
void accept_join_prune_v1(__u32 src, __u32 dst, char *p, int datalen);
void accept_join_prune(__u32 src, __u32 dst, char *p, int datalen);
int accept_assert_v1(__u32 src, __u32 dst, char *p, int datalen);
int accept_assert_v2(__u32 src, __u32 dst, char *p, int datalen);
extern void process_kernelCall(char *);
/* extern int dr(__u32 src, vifi_t); */
extern void dump_routes(FILE*);
extern struct mrtentry * longest_match_r(__u32 source, __u32 group);
void add_outgo(vifi_t vifi, struct mrtentry *r);
int encode_source_v1(char **pktAdd, struct mrtentry *mrtPtr, u_int rp_bit,
		     u_char maskLen);

/* vif.c */
extern struct vifentry viflist[MAXVIFS];
extern vifi_t numvifs;
extern vifi_t reg_vif_num;
extern vifbitmap_t nomrt_vif;
extern void init_vifs(void);
vifi_t vifi_by_index(int ifindex);
extern int igmp_ifindex;
extern void accept_group_report(__u32 src, __u32 dst, __u32 group);
void accept_query_v2(__u32 src, __u32 dst, char *p, int datalen);
void accept_query_v1(__u32 src, __u32 dst, char *p, int datalen);
extern void age_vifs(void);
extern vifi_t local_address(__u32 src);
extern void dump_vifs(FILE*);
extern void query_groups(struct vifentry *v);
struct nbrentry * neigh_lookup(vifi_t vifi, __u32 src, int version);

/* mrt.c */
struct grpentry * find_group_r(__u32 group, char create);
struct srcentry * find_source_r(__u32 src, char create);
struct mrtentry * find_route_r(__u32 src, __u32 group, u_short flags, char create);
int set_src_iif(struct srcentry *src_ent);

/* timer.c */
extern int ageRPSetTimer(void);
extern void age_routes(void);

/* inet.c */
extern char s1[];
extern char s2[];
extern char s3[];
extern char s4[];
extern char *inet_fmt(__u32, char *);
extern char *inet_fmts(__u32, __u32, char*);
int inet_valid_host(__u32 addr);
int inet_valid_subnet(__u32 nsubnet, __u32 nmask);
__u32 inet_parse(char *s);
int inet_cksum(__u16 *addr, u_int len);

#define NF_LEAF                 0x01	/* This neighbor is a leaf */
#define NF_PRUNE                0x02	/* This neighbor understands prunes */
#define NF_GENID                0x04	/* I supply genid & rtrlist in probe */
#define NF_MTRACE               0x08	/* I can understand mtrace requests */
#define NF_SNMP         0x10	/* I respond to SNMP queries */


#define MAX_DVMRP_DATA_LEN \
                ( MAX_IP_PACKET_LEN - MAX_IP_HEADER_LEN - IGMP_MINLEN )
#define VENDOR_CODE	1
#define MROUTED_VERSION	8
#define DUMMY_VERSION 8
#define MROUTED_LEVEL  ((MROUTED_VERSION << 8) | DUMMY_VERSION | \
                        ((NF_PRUNE | NF_MTRACE) << 16) | \
                        (VENDOR_CODE << 24))
#define JAN_1970 2208988800UL	/* 1970 - 1900 in seconds */
#define UNREACHABLE             32	/* "infinity" metric, must be <= 64 */

extern unsigned trace_flags;

#define TRF_IGMP	1
#define TRF_MRTSOCK	2
#define TRF_PIM1	4
#define TRF_PIM2	8
#define TRF_AUTORP	0x10
#define TRF_NETLINK	0x20
#define TRF_MRTABLE	0x40
#define TRF_VIFS	0x80
#define TRF_RPSET	0x100

#define TRF_UNUSUAL	0x80000000
#define TRF_PIM		(TRF_PIM1|TRF_PIM2)

void do_tracef (char* fmt, ...);

#define tracef(channel,msg...) do \
  if (debug && (trace_flags&(channel))) do_tracef(msg); \
  while (0)

#define tracef_flush(channel) do /* nothing */; while (0)

extern int dirty_bits;

int register_input_handler(int fd, ihfunc_t func);

void init_autorp(void);
void autorp_send_announce(void);

void accept_dvmrp(int code, __u32 src, __u32 dst);

int construct_c_rp_set(char *pkt);
int send_c_rp_set(char *packet, int datalen, __u32 src, __u32 dst);
int periodic_C_RP_Adv(void);

int accept_register_v1(__u32, __u32, char *, int);
void accept_register_stop_v1(__u32, __u32, char *, int);
int accept_rpreach_v1(__u32, __u32, char *, int);
__u32 max_local_address(void);
void dump_auto_rp_set(FILE*);

int init_routesock(void);

void autorp_vif(int ifindex, int on);

int unicast_routing_check(struct srcentry *);

char *inet_time(u_long, char *);

int pim2_encode_source(char **pktAdd, struct mrtentry *mrtPtr, u_int rp_bit, u_char maskLen);
int pim2_encode_group(char **pktAdd, __u32 group, u_char maskLen);
int pim2_encode_unicast(char **pktAdd, __u32 group);
int pim2_decode_source(char **p, void *src, u_short *flags);
int pim2_decode_group(char **p, void *grp, u_char *);
int pim2_decode_unicast(char **p, void *addr);

void send_null_register(struct mrtentry *);
void unlink_route(struct mrtentry *);
int send_rpreach_v1(struct mrtentry *, int);

void send_assert_v1(struct mrtentry *, struct vifentry*);
void send_assert_v2(struct mrtentry *mrtPtr, struct vifentry *v);

int accept_register_v2(__u32 src, __u32 dst, char *pimPkt, int pimlen);
void accept_register_stop_v2(__u32 src, __u32 dst, char *pimPkt, int pimlen);
void accept_join_prune_v2(__u32 src, __u32 dst, char *pimPkt, int pimlen);
int accept_c_rp_set_v2(__u32 src, __u32 dst, char *pimPkt, int pimlen);
int accept_rp_adv_v2(__u32 src, __u32 dst, char *pimPkt, int pimlen);

void add_to_dr_rpset(__u32 rp, __u32 grp, int masklen, int prio, u_long tmo, int vers);

int accept_register_gen(__u32 src, __u32 dst, char *pkt,
			int bb, int null_reg, void (*srs)(__u32, __u32, __u32, __u32, __u32));
void accept_reg_stop_gen(__u32 source, __u32 group);

int prune_rpt_v2(struct vifentry *v, struct nbrentry *n, struct mrtentry *mrtPtr);
int prune_rpt_v1(struct vifentry *v, struct nbrentry *n, struct mrtentry *mrtPtr);
void send_null_register_v2(struct mrtentry *r);
void send_null_register_v1(struct mrtentry *r);
int accept_assert_gen(vifi_t vifi, struct nbrentry *n,
		      __u32 source, __u32 group,
		      __u32 preference, __u32 metric);

void do_accept_prune(vifi_t vifi, __u32 source, __u32 group, u_short flags);
void do_accept_join(vifi_t vifi, __u32 source, __u32 group, u_short flags);

int event_join_prune_v1(struct mrtentry *r, char join,
			struct nbrentry *n, struct grpentry *g,
			struct vifentry *v);
int event_join_prune_v2(struct mrtentry *r, char join,
			struct nbrentry *n, struct grpentry *g,
			struct vifentry *v);
int periodic_join_prune_v1(struct nbrentry *n, vifi_t vifi);
int periodic_join_prune_v2(struct nbrentry *n, vifi_t vifi);
void query_neighbors_v2(struct vifentry*);
void query_neighbors_v1(struct vifentry*);

int is_local_on(__u32 target, struct vifentry *v);

#define JOIN 1
#define PRUNE 2
#define NONE 3

int join_or_prune(struct nbrentry *n, vifi_t vifi, struct mrtentry *r);

void send_register_v1(__u32 dest, struct ip *ip, int pktlen);
void send_register_v2(__u32 dest, struct ip *ip, int pktlen);
int initBSR(void);
