







/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: inet.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */
#include "defs.h"
#include <stdlib.h>
char s1[16];
char s2[16];
char s3[16];
char s4[16];

void dump_oif(long oifs, char *msg)
{
	int vifi;
	log(LOG_DEBUG, 0, "dump_oif: %s", msg);
	for (vifi = 0; vifi < numvifs; vifi++) {
		if (VIFM_ISSET(vifi, oifs)) {
			log(LOG_DEBUG, 0, "   Out Vif: eth%d", vifi);
		}
	}



}

int inet_valid_host(__u32 addr)
{
	addr = ntohl(addr);
	return (!(IN_MULTICAST(addr) || IN_BADCLASS(addr) || (addr & 0xff000000) == 0));
}

int inet_valid_subnet(__u32 nsubnet, __u32 nmask)
{
	__u32 subnet, mask;

	subnet = ntohl(nsubnet);
	mask = ntohl(nmask);

	if ((subnet & mask) != subnet)
		return (FALSE);

	if (mask == 0 || subnet == 0)
		return FALSE;

	return (TRUE);
}


char * inet_fmt(__u32 addr, char *s)
{
	register u_char *a;

	a = (u_char *) & addr;
	sprintf(s, "%u.%u.%u.%u", a[0], a[1], a[2], a[3]);
	return (s);
}

char * inet_fmts(__u32 addr, __u32 mask, char *s)
{
	register u_char *a, *m;

	a = (u_char *) & addr;
	m = (u_char *) & mask;

	if (m[3] != 0)
		sprintf(s, "%u.%u.%u.%u", a[0], a[1], a[2], a[3]);
	else if (m[2] != 0)
		sprintf(s, "%u.%u.%u", a[0], a[1], a[2]);
	else if (m[1] != 0)
		sprintf(s, "%u.%u", a[0], a[1]);
	else
		sprintf(s, "%u", a[0]);

	return (s);
}


__u32 inet_parse(char *s)
{
	u_long a;
	u_int a0, a1, a2, a3;
	char c;

	if (sscanf(s, "%u.%u.%u.%u%c", &a0, &a1, &a2, &a3, &c) != 4 ||
	    a0 > 255 || a1 > 255 || a2 > 255 || a3 > 255)
		return (0xffffffff);

	((u_char *) & a)[0] = a0;
	((u_char *) & a)[1] = a1;
	((u_char *) & a)[2] = a2;
	((u_char *) & a)[3] = a3;

	return (a);
}


int inet_cksum(__u16 *addr, u_int len)
{
	register int nleft = (int) len;
	register u_short *w = addr;
	u_short answer = 0;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while (nleft > 1) {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(u_char *) (&answer) = *(u_char *) w;
		sum += answer;
	}
	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);	/* add carry */
	answer = ~sum;		/* truncate to 16 bits */
	return (answer);
}

char *inet_time(unsigned long val, char *buffer)
{
	time_t now;
	struct tm t;
	div_t value1, value2;
	double secs;
	now = wall_time;
	secs = difftime(now, val);
	if (secs < 0)
		secs = -secs;
	value1 = div(secs, 3600);
	t.tm_hour = value1.quot;
	value2 = div(value1.rem, 60);
	t.tm_min = value2.quot;
	t.tm_sec = value2.rem;
	sprintf(buffer, "%02d:%02d:%02d", t.tm_hour, t.tm_min, t.tm_sec);
	return buffer;
}
