/* Copyright (c) 1996 University of Southern CaSlifornia
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: kern.c
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */
#include "defs.h"

u_int numberOfCaches = 0;
u_int numberOfAddMfcs = 0;

void k_igmp_hdr_include(int b)
{
#ifdef IP_HDRINCL
	if (setsockopt(igmp_socket, IPPROTO_IP, IP_HDRINCL,
		       (char *) &b, sizeof(b)) < 0)
		log(LOG_ERR, errno, "setsockopt IP_HDRINCL %u", b);
#endif	/* IP_HDRINCL */
}


void k_igmp_set_rcvbuf(int size)
{
	if (setsockopt(igmp_socket, SOL_SOCKET, SO_RCVBUF,
		       (char *) &size, sizeof(size)) < 0)
		log(LOG_ERR, errno, " setsockopt SO_RCVBUF %u", size);
}

void k_igmp_set_loop(u_char b)
{
	static int igmp_loop = -1;
	if (igmp_loop != b) {
		if (setsockopt(igmp_socket, IPPROTO_IP, IP_MULTICAST_LOOP,
			       (char *) &b, sizeof(b)) < 0)
			log(LOG_ERR, errno, "setsockopt IP_MULTICAST_LOOP %u", b);
		igmp_loop = b;
	}
}

void k_pim_hdr_include(b)
int b;
{
#ifdef IP_HDRINCL
	if (setsockopt(pim_socket, IPPROTO_IP, IP_HDRINCL,
		       (char *) &b, sizeof(b)) < 0)
		log(LOG_ERR, errno, "setsockopt IP_HDRINCL %u", b);
#endif	/* IP_HDRINCL */
}


void k_pim_set_rcvbuf(size)
int size;
{
	if (setsockopt(pim_socket, SOL_SOCKET, SO_RCVBUF,
		       (char *) &size, sizeof(size)) < 0)
		log(LOG_ERR, errno, " setsockopt SO_RCVBUF %u", size);
}

void k_pim_set_loop(b)
u_char b;
{
	static int pim_loop = -1;
	if (pim_loop != b) {
		if (setsockopt(pim_socket, IPPROTO_IP, IP_MULTICAST_LOOP,
			       (char *) &b, sizeof(b)) < 0)
			log(LOG_ERR, errno, "setsockopt IP_MULTICAST_LOOP %u", b);
		pim_loop = b;
	}
}

void k_pim_set_rxinfo(u_char t)
{
	int on = t;
	if (setsockopt(pim_socket, IPPROTO_IP, IP_PKTINFO,
		       (char *) &on, sizeof(on)) < 0)
		log(LOG_ERR, errno, "setsockopt IP_PKTINFO %u", t);
}

void k_igmp_set_rxinfo(u_char t)
{
	int on = t;
	if (setsockopt(igmp_socket, IPPROTO_IP, IP_PKTINFO,
		       (char *) &on, sizeof(on)) < 0)
		log(LOG_ERR, errno, "setsockopt IP_PKTINFO %u", t);
}

void k_init_pim()
{
	int v = 1;
	if (debug > 5)
		log(LOG_DEBUG, 0, "MRT_INIT on igmp_socket\n");
	if (setsockopt(igmp_socket, IPPROTO_IP, MRT_INIT,
		       (char *) &v, sizeof(int)) < 0)
		 log(LOG_ERR, errno, "can't enable PIM routing in kernel");
	if (setsockopt(igmp_socket, IPPROTO_IP, MRT_PIM,
		       (char *) &v, sizeof(int)) < 0)
		 log(LOG_ERR, errno, "can't set PIM flag in kernel");
}


void k_stop_pim()
{
	int v = 0;
	if (debug > 9)
		log(LOG_DEBUG, 0, "MRT_DONE on igmp_socket\n");
	if (setsockopt(igmp_socket, IPPROTO_IP, MRT_DONE,
		       (char *) NULL, 0) < 0)
		log(LOG_WARNING, errno, "can't disable PIM routing in kernel");
	if (debug > 9)
		log(LOG_DEBUG, 0, "MRT_PIM on igmp_socket\n");

	if (setsockopt(igmp_socket, IPPROTO_IP, MRT_PIM,
		       (char *) &v, sizeof(int)) < 0)
		 log(LOG_ERR, errno, "can't reset PIM flag in kernel");
}

void k_add_vif(vifi)
vifi_t vifi;
{

	register struct vifentry *v;
	struct vifctl vc;
	memset(&vc, 0, sizeof(struct vifctl));
	v = &viflist[vifi];
	vc.vifc_vifi = vifi;
	vc.vifc_flags = v->flags;
	vc.vifc_threshold = v->threshold;
	vc.vifc_lcl_addr.s_addr = v->prefsrc;
	vc.vifc_rate_limit = v->rate_limit;

	if (debug > 9)
		log(LOG_DEBUG, 0, "MRT_ADD_VIF on igmp_socket: %s %d %s\n",
		    v->name, vifi, v->name);

	if (setsockopt(igmp_socket, IPPROTO_IP, MRT_ADD_VIF,
		       (char *) &vc, sizeof(vc)) < 0)
		log(LOG_ERR, errno, "setsockopt PIM_ADD_VIF: fl=%02x", vc.vifc_flags);
}


void k_del_vif(vifi)
vifi_t vifi;
{

	register struct vifentry *v;
	struct vifctl vc;
	memset(&vc, 0, sizeof(struct vifctl));
	v = &viflist[vifi];
	vc.vifc_vifi = vifi;
	vc.vifc_flags = v->flags;
	vc.vifc_threshold = v->threshold;
	vc.vifc_lcl_addr.s_addr = v->prefsrc;

	if (debug > 9)
		log(LOG_DEBUG, 0, "MRT_DEL_VIF on igmp_socket\n");

	if (setsockopt(igmp_socket, IPPROTO_IP, MRT_DEL_VIF,
		       (char *) &vc, sizeof(vc)) < 0)
		log(LOG_ERR, errno, "setsockopt PIM_DEL_VIF");
}


void k_chg_viff(vifi)
vifi_t vifi;
{

	register struct vifentry *v;
	struct vifctl vc;
	memset(&vc, 0, sizeof(struct vifctl));
	v = &viflist[vifi];
	vc.vifc_vifi = vifi;
	vc.vifc_flags = v->flags;
	vc.vifc_threshold = v->threshold;
	vc.vifc_lcl_addr.s_addr = v->prefsrc;

	if (debug > 9)
		log(LOG_DEBUG, 0, "MRT_ADD_VIF on igmp_socket\n");

	if (setsockopt(igmp_socket, IPPROTO_IP, MRT_ADD_VIF,
		       (char *) &vc, sizeof(vc)) < 0)
		log(LOG_WARNING, errno, "setsockopt PIM_CHG_VIFF");

}

/* function name: k_del_mrt
   input: mrt entry pointer
   output: int (TRUE if successful, FALSE o.w.)
   operation: deletes the given mrt from the kernel cache
   comments: called by delete_mrt [route.c]
 */
int k_del_mrt(r)
struct mrtentry *r;
{
	struct mfcctl mc;
	memset(&mc, 0, sizeof(struct mfcctl));
	if (!(r->flags & MRTF_KERNEL_CACHE)) {
		if (debug > 1)
			log(LOG_DEBUG, 0, "\nDelete mfc: not kernel cache");
		return FALSE;
	}
	mc.mfcc_origin.s_addr = r->source->source;
	mc.mfcc_mcastgrp.s_addr = r->group->group;

	if (debug > 1)
		log(LOG_DEBUG, 0, "\nDeleting mfc, src %s, grp %s",
		    inet_fmt(r->source->source, s1),
		    inet_fmt(r->group->group, s2));

	if (setsockopt(igmp_socket, IPPROTO_IP, MRT_DEL_MFC, (char *) &mc,
		       sizeof(mc)) < 0) {
		log(LOG_WARNING, errno, "setsockopt k_del_mrt");
		return FALSE;
	}
	numberOfCaches--;

	/* turn off the cache flag */
	r->flags &= ~MRTF_KERNEL_CACHE;
	return (TRUE);
}

/* function name: k_chg_mrtf
   input: mrtentry pointer
   output: int (TRUE if successful, FALSE o.w.)
   operation: installs/modifies a cache entry in the kernel
   comments: called by update_kernel_leaves, update_kernel_oifs,
   accept_register, process_cacheMiss, and process_wrongiif
 */
int k_chg_mrtf(struct mrtentry *r)
{
	struct mfcctl mc;
	vifbitmap_t oifs;
	struct vifentry *v;
	vifi_t vifi;
	struct mrtentry *rp = NULL;

	memset(&mc, 0, sizeof(struct mfcctl));
	mc.mfcc_origin.s_addr = r->source->source;
	mc.mfcc_mcastgrp.s_addr = r->group->group;

	if (debug > 3)
		log(LOG_DEBUG, 0, "k_chg_mrtf: src %s, grp %s",
		    inet_fmt(r->source->source, s1),
		    inet_fmt(r->group->group, s2));

	/* oif = (outgoing - maskoff) + leaves */
	oifs = 0;		/* initialization */
	if (!(r->flags & MRTF_NULL_OIF)) {
		if ((rp = r->group->rp_entry) != NULL)
			oifs = (rp->outgoing & ~(r->vifmaskoff)) & ~nomrt_vif;

		oifs |= ((r->outgoing & ~nomrt_vif) | r->group->leaves);
	}			/* if not 'NULL OIF' */
	/* set the parent of the entry */
	if (r->flags & MRTF_IIF_REGISTER) {
		mc.mfcc_parent = reg_vif_num;
		if (debug > 4)
			log(LOG_DEBUG, 0, "\n IIF_REGISTER flag set, iif = %d",
			    reg_vif_num);
	} else {
		if (rp && !(r->flags & MRTF_SPT)) {	/* if the S,G is on the shared tree and the spt is cleared,
							 * then we RPF towards the RP.
							 */
			if (!r->source->upstream || (local_address(rp->source->source) != MAXVIFS)) {
				/* I am the RP for a directly connected source or just the router for the directly conneced source ??? FIX  */
				mc.mfcc_parent = r->source->incoming;
			} else {
				if (debug > 3)
					log(LOG_DEBUG, 0, "Not Found %s (rp->source->source)  in local address, changing to rp incoming",
					inet_fmt(rp->source->source, s1));
				mc.mfcc_parent = rp->source->incoming;
			}
		} else
			mc.mfcc_parent = r->source->incoming;

	}

	/*  exclude the iif from the oifs */
	if (debug > 2)
		log(LOG_DEBUG, 0, "Excluding %d from (%s,%s) oif list",
		    mc.mfcc_parent,
		    inet_fmt(r->source->source, s1),
		    inet_fmt(r->group->group, s2));

	VIFM_CLR(mc.mfcc_parent, oifs);


	if (mc.mfcc_parent > numvifs) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "iif = %u, invalid", mc.mfcc_parent);
		return (FALSE);
	}
	if (debug > 3)
		log(LOG_DEBUG, 0, "    In Vif: %d", mc.mfcc_parent);

	for (vifi = 0, v = viflist; vifi < numvifs; vifi++, v++) {
		if (VIFM_ISSET(vifi, oifs)) {
			mc.mfcc_ttls[vifi] = v->threshold;
			if (debug > 3)
				log(LOG_DEBUG, 0, "   Out Vif: %d, with ttl: %d",
				    vifi,
				    v->threshold);
		} else
			mc.mfcc_ttls[vifi] = 0;
	}

	/* chk REGISTER and add reg_vif_num to oif if set. */
	if (r->flags & MRTF_REGISTER) {
		mc.mfcc_ttls[reg_vif_num] = viflist[reg_vif_num].threshold;
		if (debug > 4)
			log(LOG_DEBUG, 0, "   Out Vif: %d (pim_reg), with ttl: %d",
			    reg_vif_num,
			    viflist[reg_vif_num].threshold);
	} else {		/* if not a register entry */
		mc.mfcc_ttls[reg_vif_num] = 0;
	}			/* if not register */

	if (!(r->flags & MRTF_KERNEL_CACHE)) {
		r->flags |= MRTF_KERNEL_CACHE;
		dirty_bits |= 2;
	}

	if (r->flags & MRTF_CRT) {
		r->flags &= ~MRTF_CRT;
		dirty_bits |= 2;
	}

	numberOfAddMfcs++;
	if (!oifs && !(r->flags & MRTF_REGISTER) && (debug > 4))
		log(LOG_DEBUG, 0, "   Out Vif: NULL");

	r->outgoing = oifs;
	if (setsockopt(igmp_socket, IPPROTO_IP, MRT_ADD_MFC, (char *) &mc,
		       sizeof(mc)) < 0) {
		log(LOG_WARNING, errno, "setsockopt k_chg_mrt");
		return (FALSE);
	}
	return (TRUE);
}

u_long k_get_sg(struct mrtentry *mrt)
{
	struct sioc_sg_req req;

	req.src.s_addr = mrt->source->source;
	req.grp.s_addr = mrt->group->group;
	if (debug > 9)
		log(LOG_DEBUG, 0, "SIOCGETSGCNT on igmp_socket\n");

	if (ioctl(igmp_socket, SIOCGETSGCNT, (char *) &req) < 0) {
		log(LOG_WARNING, errno, "ioctl SIOCGETSGCNT (S=%s,G=%s)",
		    inet_fmt(req.src.s_addr, s1),
		    inet_fmt(req.grp.s_addr, s2));
		return -1;
	}
	return req.pktcnt;
}

void k_joinleave(int so, struct vifentry *v, __u32 mc, int joinleave)
{
	struct ip_mreqn imr;	/* multicast request struct to join the group */
	int op = joinleave ? IP_ADD_MEMBERSHIP : IP_DROP_MEMBERSHIP;

	memset(&imr, 0, sizeof(struct ip_mreqn));
	imr.imr_multiaddr.s_addr = mc;
	imr.imr_ifindex = v->ifindex;
	if (setsockopt(so, IPPROTO_IP, op, &imr, sizeof(imr)) < 0) {
		log(LOG_DEBUG, errno,
		    "%s, ifindex: %d  can't join (%s)",
		    v->name,
		    imr.imr_ifindex,
		    inet_fmt(mc, s1));
	} else {
		if (debug > 1)
			log(LOG_NOTICE, 0,
			    "%s joined on interface: %s, ifindex %d",
			    inet_fmt(mc, s1),
			    v->name,
			    v->ifindex);
	}
}
