
/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
/* 
 * PIMv1 Linux Version: Copyright (c) 1997 University of Lisbon 
 * All rights reserved.
 *
 * FILE: pim.h
 * Changes:
 * Carlos.Picoto@di.fc.ul.pt           Linux Changes
 *
 */

#define INADDR_ALLROUTERS_GROUP (__u32)0xe0000002	/* 224.0.0.2   */
#define INADDR_ALL_PIM_ROUTERS (__u32)0xe000000D     /*224.0.0.13 */

#define TRUE    		1
#define FALSE   		0

#define BUFSIZE      48*1024
#define MINTTL       1
#define DEBUG        2

#define EMPTY_VIF	0x00000000
#define ALL_VIF		0xffffffff
#define EMPTY_FLAG	0x0000

#define SRC_MSKLEN 32		/* the source masklength now is 32 */
#define GRP_MSKLEN 32		/* the group mask length is 32 */
#define HASHMASKLEN 30		/* default to hash mask length of 30 */

#define VIFF_DR                 0x0100	/* designated router    */
#define VIFF_NOMRT              0x0200	/* no neighbor on vif   */
#define VIFF_DOWN               0x0400	/* interface is down    */
#define VIFF_DISABLED           0x0800	/* disabled interafce   */

#define MRTF_SPT                0x0001	/* on shortest path tree  */
#define MRTF_WC                 0x0002	/* wildcard entry         */
#define MRTF_RP                 0x0004	/* incoming vif toward RP */
#define MRTF_CRT                0x0008	/* create by kernel       */
#define MRTF_IIF_REGISTER	0x0020
#define MRTF_REGISTER		0x0080
#define MRTF_KERNEL_CACHE	0x0200	/* a mirror for the kernel cache */
#define MRTF_NULL_OIF		0x0400	/* null oif cache.. */
#define MRTF_REG_SUPP		0x0800	/* register suppress */
#define MRTF_ASSERTED		0x1000	/* upstream is not that of src */
#define MRTF_SG			0x2000	/* (S,G) pure, not hanging off of (*,G) */
#define MRTF_DENSE		0x4000

#define MAX_IP_PACKET_LEN 	4096	/*Made 4k as this creates problems..576 */
#define MIN_IP_HEADER_LEN 	20
#define MAX_IP_HEADER_LEN 	60
#define PIM_MINLEN		4
#define IGMP_HDR_LEN            4
#define MAX_PIM_DATA_LEN  	(MAX_IP_PACKET_LEN  - MAX_IP_HEADER_LEN  - PIM_MINLEN)

#define ADDRSIZE 6		/* Encoded Address Size for join prune. */
#define ADDR_LEN 4		/* Size of the unicast address */
#define MAXSrc 20		/* Max number of sources */

#define TIMER_INTERVAL		        5
#define HelloPeriod 			30
#define GROUP_QUERY_INTERVAL		60
#define JOIN_PRUNE_INTERVAL		60
#define C_RP_Adv_INTERVAL		60	/* 1 min */
#define BootstrapPeriod			60
#define BootstrapTimeout	(2.5 * BootstrapPeriod)
#define HelloHoldtime        	(3.5 * HelloPeriod)
#define LEAF_TIMEOUT            (3.5 * GROUP_QUERY_INTERVAL)
#define AssertTimeout           180
#define OUTGOING_TIMEOUT	(3.5 * JOIN_PRUNE_INTERVAL)
#define ROUTE_TIMEOUT		OUTGOING_TIMEOUT
#define RPHoldtime		(2.5 * C_RP_Adv_INTERVAL)
#define REGISTER_BIT_TIMEOUT	30	/* 60 */
#define REGISTER_RATE_TIMER	5	/* check reg rate after this interval */
#define DATA_RATE_TIMER		30	/* check data rate after this interval */

#define PIM_HELLO               0
#define PIM_REGISTER            1
#define PIM_REGISTER_STOP       2
#define PIM_JOIN_PRUNE          3
#define PIM_BOOTSTRAP		4
#define PIM_ASSERT              5
#define PIM_GRAFT               6
#define PIM_GRAFT_ACK           7
#define PIM_CANDIDATE_RP_ADVERTISEMENT 8

#define PIMv1_QUERY               0
#define PIMv1_REGISTER            1
#define PIMv1_REGISTER_STOP       2
#define PIMv1_JOIN_PRUNE          3
#define PIMv1_RPREACHABILITY      4
#define PIMv1_ASSERT              5
#define PIMv1_GRAFT               6
#define PIMv1_GRAFT_ACK           7

#define MAX_RATE_LIMIT          100000	/* max rate limit               */
#define DEFAULT_PHY_RATE_LIMIT  0	/* default phyint rate limit    */

struct vifaddr
{
	struct vifaddr *next;
	__u32 local;
	__u32 prefix;
	__u32 mask;
	int primary;
};

struct vifentry {
	u_short flags;		/* VIFF_ flags                  */
	u_char threshold;	/* min ttl required             */
	__u32 prefsrc;
	struct vifaddr *ifa;
	char name[IFNAMSIZ];	/* interface name               */
	u_char timer;		/* timer for sending queries    */
	u_short query_holdtime;	/* the holdtime for queries     */
	u_char gq_timer;	/* Group Query timer, used by DR */
	struct nbrentry *neighbors;	/* list of neighboring routers  */
	int		v1neighnum;	/* Number of pure V1 neighbors */
	int		v2neighnum;     /* Number of pure V2 neighbors */
	int		neighnum;	/* Total number neighbors */
	u_int rate_limit;	/* max rate                     */
	int ifindex;		/* Interface index              */
	__u32 metric;
};

struct nbrentry {
	__u32 address;		/* neighbor address             */
	vifi_t vif;		/* which interface              */
	u_short timer;		/* for timing out neighbor      */
	u_short holdtime;	/* the holdtime, from queries   */
	u_short timer2;		/* for timing out neighbor      */
	u_short holdtime2;	/* the holdtime, from queries   */
	u_char mode;
	u_char pimversion;
	u_long uptime;
	struct nbrentry *next;	/* link to next neighbor        */
};


struct srcentry {
	unsigned hash;
	__u32 source;		/* subnet source of multicasts  */
	vifi_t incoming;	/* incoming vif                 */
	struct nbrentry *upstream;	/* upstream router              */
	u_short timer;		/* timer to recompute incoming  */
	u_char asserted;
	u_char direct;
	__u32 metric;		/* Unicast Routing Metric for the source */
	__u32 preference;	/* The metric preference value */
	u_long uptime;
	struct mrtentry *mrtlink;	/* link to routing entries      */
	struct srcentry *next;	/* link to next entry           */
};

#define SRCHASH_SIZE 256

#define forall_sources(s) { \
  int __h_idx; \
  for (__h_idx = 0; __h_idx < SRCHASH_SIZE; __h_idx++) { \
    for ((s) = srchash[__h_idx]; (s); (s) = (s)->next)

#define endforall_sources(s) } }


/* Struture for list of rps for a group */
struct rplist {
	struct rplist *next;
	__u32 rpaddr;
	unsigned char pimversion;
	u_short timer;		/* Status Timer */
};

struct g_rplist {
	struct g_rplist *next;
	__u32 rpaddr;
	u_short timer;		/* Status Timer */
	unsigned char grpmasklen;
	unsigned char pimversion;
	unsigned char dense;
	unsigned char announce;
	unsigned char scope;
	unsigned char override;
	__u32 info_source;
	u_long  expires;
	__u32  group;
	__u32  grpmask;
};

struct rp_afilter
{
	struct rp_afilter *next;
	__u32 group;
	__u32 grpmask;
	int   grplen;
	__u32 rpaddr;
	__u32 rpmask;
	__u32 oaddr;
	__u32 omask;
	unsigned flag;
};

extern struct rp_afilter *rp_discovery_filter;

struct mrt_static
{
	struct mrt_static *next;
	__u32 addr;
	__u32 mask;
	int   ifindex;
};

extern struct mrt_static *mrt_static_routes;


struct grpentry {
	unsigned hash;
	__u32 group;		/* subnet group of multicasts     */
	vifbitmap_t leaves;	/* outgoing vif to host           */
	struct mrtentry *mrtlink;	/* link to routing entries        */
	struct grpentry *next;	/* link to next entry             */
	u_char flags;		/* Whether the group is DENSE/SPARSE */
	struct mrtentry *rp_entry;	/* Pointer to the *,g entry */
	struct rplist *active_rp;	/* Pointer to the active RP */
	u_long uptime;
	u_char timers[0];
};

#define GRPHASH_SIZE 256

#define forall_groups(g) { \
  int __h_idx; \
  for (__h_idx = 0; __h_idx < GRPHASH_SIZE; __h_idx++) { \
    for ((g) = grphash[__h_idx]; (g); (g) = (g)->next)

#define endforall_groups(g) } }



struct mrtentry {
	unsigned hash;
	struct srcentry *source;	/* source                       */
	struct grpentry *group;	/* group                        */
	vifbitmap_t outgoing;	/* outgoing vifs to downstream  */
	vifbitmap_t vifmaskoff;	/* deleted vifs                 */
	struct mrtentry *next;		/* next entry of same hash	*/
	struct mrtentry *srcnext;	/* next entry of same source    */
	struct mrtentry *grpnext;	/* next entry of same group     */
	struct nbrentry *upstream;	/* upstream router, needed as in
					 * RPbit entry, the upstream
					 * router is different than
					 * the source upstream router. 
					 *
					 * BEWARE! This comment is wrong,
					 * r->upstream is used only in asserted
					 * entries --ANK
					 */
	__u32 preference;
	__u32 metric;
#define DATA_RATE_LIMIT 300
	u_long pktrate_prev;	/* packet count of prev check   */
	u_long idlecnt_prev;	/* pkt cnt to check idle states */
	u_int data_rate_timer;	/* to keep track of the data rate */
	u_int reg_rate_timer;	/* to keep track of register rate at RP */
#define REGISTER_RATE_LIMIT 1	/* Original 25 */
	u_int reg_cnt;		/* to keep track of the register count at RP */
#define SEND_REGISTER_RATE_LIMIT 10
	u_int reg_snd_cnt;	/* Keep track of register sents per timer interval */
	u_short timer;		/* entry timer                  */
	u_short join_prune_timer;	/* timer for periodic join_prune */
	u_short flags;		/* flags, now short as more flags needed */
	u_short jtimer;		/* The joiner bit timer         */
#define ASSERT_RPF_TIMER 90
	u_int assert_rpf_timer;
#define ASSERT_RATE_TIMER 15
	u_int assert_rate_timer;
	u_short registerBitTimer;	/* the register-bit timer */
	u_short rpreachtimer;
	u_long  __align_boundary;
	u_char  timers[0];
};

#define MRTHASH_SIZE 256

/*
 * PIM packet header, length is 4 bytes
 */
struct pimhead {
	__u8	pim_vertype;
	__u8	pim_reserved;
	__u16	pim_cksum;
};


/*
 * structure used by PIM to get the RPF neighbor and IIF info
 * for a given source, from the unicast routing tables.
 * Maybe be substituted by routing sockets for the supporting systems
 * credits to Ahmed Helmy / SGI
 */

struct rpfctl {
	struct in_addr source;	/* the source for which we want iif and rpfnbr */
	struct in_addr rpfneighbor;	/* next hop towards the source */
	vifi_t iif;		/* the incoming interface to reach the next hop */
};

struct mga_join_prune {
	__u32 gaddr;
	__u32 gaddrmask;
	__u16 njoins;
	__u16 nprunes;
	/* encoded addr list: it is impossible to make in C because
	   of wrong protocol alignment.
	 */
	char  eaddrs[0];
};

struct pimquery {
	__u8 mode;
	__u8 reserved;
	__u16 holdtime;
};

#define PIMv1_QUERY_SIZE 4
struct pimreg_stop {
	__u32 gaddr;
	__u32 saddr;
};

struct pim_jp {
	__u32 up_neigh_addr;
	__u16 reserved;
	__u16 holdtime;
	__u8 reserved2;
	__u8 maddrlen;
	__u8 addrlen;
	__u8 numgroups;
	struct mga_join_prune list[0];
};

struct pim_reachability {
	__u32 gaddr;
	__u32 gmask;
	__u32 rpaddr;
	__u16 reserved;
	__u16 holdtime;
};

struct pim_assert {
	__u32 gaddr;
	__u32 gmask;
	__u32 saddr;
	__u32 metricpref;
	__u32 metric;
};
