#include "defs.h"

u_char jpholdtime[2] =
{0, OUTGOING_TIMEOUT};

int encode_source_v1(char **pktAdd, struct mrtentry *mrtPtr, u_int rp_bit,
		     u_char maskLen)
{
	int datalen = 0;
	u_char flags1, flags2;
	flags1 = *((*pktAdd)++) = 1;
	datalen++;
	flags2 = *((*pktAdd)++) =
		(((mrtPtr->flags & MRTF_WC) ? 1 : 0) << 7) |
		((((mrtPtr->flags & MRTF_RP) | rp_bit) ? 1 : 0) << 6) |
		32;
	if (rp_bit | (mrtPtr->flags & MRTF_WC))
		log(LOG_DEBUG, 0, "PIM:Sending WC Join (*,%s) fl1:%x, fl2:%x",
		    inet_fmt(mrtPtr->group->group, s1),
		    flags1,
		    flags2);

	datalen++;
	bcopy((char *) &(mrtPtr->source->source), *pktAdd, 4);
	datalen += 4;
	(*pktAdd) += 4;
	return (datalen);
}

void query_neighbors_v1(struct vifentry *v)
{
	struct pimquery *pim1;
	struct vifaddr *vifa;

	if (v->flags & VIFF_DOWN)
		return;

	pim1 = (struct pimquery *)
		(pimv1_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN);
	pim1->mode = 2<<4;
	pim1->reserved = 0;
	pim1->holdtime = htons(v->query_holdtime);

	tracef(TRF_PIM1, "query_neighbors_v1 on %s Holdtime=%d",
	       v->name,
	       v->query_holdtime);

	v->timer = 0;
	for (vifa = v->ifa; vifa; vifa = vifa->next) {
		if (!vifa->primary)
			continue;
		send_pim_v1(pimv1_send_buf,
			    vifa->local,	
			    allrouters_group,
			    PIMv1_QUERY, PIMv1_QUERY_SIZE, v->ifindex, FALSE, 1);
	}
}

static void send_reg_stop_v1(__u32 source, __u32 group, __u32 regsrc,
			     __u32 regdst, __u32 active_rp)
{
	struct pimreg_stop *prs = (void*)(pimv1_send_buf + MIN_IP_HEADER_LEN + PIM_MINLEN);

        prs->gaddr = group;
        prs->saddr = source;
	send_pim_v1(pimv1_send_buf, regsrc, regdst,
		    PIMv1_REGISTER_STOP, sizeof(*prs), 0, -1, MAXTTL);
}


int accept_register_v1(__u32 src, __u32 dst, char *pkt, int pimlen)
{
	if (pimlen < sizeof(struct ip)) {
		log(LOG_DEBUG, 0, "\ntruncated IP in register");
		return FALSE;
	}

	return accept_register_gen(src, dst, pkt, 0,
				   pimlen == sizeof(struct ip), send_reg_stop_v1);
}

void accept_register_stop_v1(__u32 src, __u32 dst, char *p, int datalen)
{
	struct pimreg_stop *pim1;

	pim1 = (struct pimreg_stop *) (p);

	tracef(TRF_PIM1, "accept_register_stop_v1: (%s,%s)",
	       inet_fmt(pim1->saddr, s1), inet_fmt(pim1->gaddr, s2));

	accept_reg_stop_gen(pim1->saddr, pim1->gaddr);
}


void accept_register_stop_v2(__u32 src, __u32 dst, char *p, int datalen)
{
	__u32 source, group;

	if (pim2_decode_group(&p, &group, NULL) == FALSE ||
	    pim2_decode_unicast(&p, &source) == FALSE)
		return;
	accept_reg_stop_gen(source, group);
}

int send_rpreach_v1(struct mrtentry *r, int holdtime)
{
	int pktlen = 0;
	vifi_t vifi;
	struct pim_reachability *pr;

	pr = (void*)(pimv1_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN);
	pr->gaddr = r->group->group;
	pr->gmask = 0xFFFFFFFF;
	pr->rpaddr = r->group->active_rp->rpaddr;
	pr->reserved = 0;
	pr->holdtime = htons(holdtime);

	pktlen = sizeof(*pr);

	for (vifi=0; vifi<numvifs; vifi++) {
		if (VIFM_ISSET(vifi, r->outgoing) &&
		    viflist[vifi].v2neighnum != viflist[vifi].neighnum) {
			if (debug)
				log(LOG_DEBUG, 0, "send_rpreach_v1 for (*,%s), rp=%s via %s",
				    inet_fmt(pr->gaddr, s1),
				    inet_fmt(pr->rpaddr, s2),
				    viflist[vifi].name);
			send_pim_v1(pimv1_send_buf,
				    viflist[vifi].prefsrc,
				    allrouters_group,
				    PIMv1_RPREACHABILITY,
				    pktlen,
				    viflist[vifi].ifindex, FALSE, 1);
		}
	}
	return (TRUE);
}

int accept_rpreach_v1(__u32 src, __u32 dst, char *p, int datalen)
{
	vifi_t vifi;
	vifi_t iif;
	struct srcentry *s;
	struct grpentry *g;
	struct mrtentry *r;
	struct pim_reachability * pr = (struct pim_reachability *)p;

	tracef(TRF_PIM1, "accept_rpreach_v1: src=%s; grp=%s/%s, rp=%s, ht=%u",
	       inet_fmt(src, s1),
	       inet_fmt(pr->gaddr, s2),
	       inet_fmt(pr->gmask, s3),
	       inet_fmt(pr->rpaddr, s4),
	       ntohs(pr->holdtime));

	if (pr->gmask != 0xFFFFFFFF) {
		tracef(TRF_UNUSUAL|TRF_PIM1, "accept_rpreach_v1: mask is %s",
		       inet_fmt(pr->gmask, s1));
		return FALSE;
	}

	/* get the interface which received the message */

	/* Because of bug in Cisco IOS (it preserves originator
	   address as IP source!!!), we cannot check for locality
	   of sender --ANK */
	if ((vifi = vifi_by_index(igmp_ifindex)) == MAXVIFS) {
		log(LOG_DEBUG, 0, "accept_rpreach_v1: ignoring RP-reachability from router %s: if%d",
			    inet_fmt(src, s1), igmp_ifindex);
		return FALSE;
	}

	s = find_source_r(pr->rpaddr, TRUE);
	if (s == NULL || s->incoming != vifi)
		return FALSE;

	if (!(g = find_group_r(pr->gaddr, FALSE))) {
		tracef(TRF_UNUSUAL, "no group");
		return FALSE;
	}
	
	if (!(r = find_route_r(g->active_rp->rpaddr, pr->gaddr,
			       MRTF_WC | MRTF_RP, FALSE))) {
		tracef(TRF_UNUSUAL, "no mroute");
		return (FALSE);
	}

	if (r->flags & MRTF_ASSERTED)
		iif = r->upstream->vif;
	else
		iif = r->source->incoming;

	if (iif == vifi) {
		if (ntohl(pr->rpaddr) > ntohl(r->group->active_rp->rpaddr)) {
			/* set rp to pr->rpaddr */
			;
		} else if (pr->rpaddr == r->group->active_rp->rpaddr) {
			/* reset timer */
			send_rpreach_v1(r, ntohs(pr->holdtime));
		}
	} else if (r->flags & MRTF_IIF_REGISTER) {
		if (ntohl(pr->rpaddr) > ntohl(r->group->active_rp->rpaddr)) {
			/* set rp to pr->rpaddr */
			/* set iif to vifi */
			send_rpreach_v1(r, ntohs(pr->holdtime));
			;
		}
	}

	return (TRUE);
}


int prune_rpt_v1(struct vifentry *v, struct nbrentry *n,
		 struct mrtentry *mrtPtr)
{
	struct pim_jp *pjp;
	struct mga_join_prune *ma;
	char *cp;

	pjp = (struct pim_jp*)(pimv1_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN);
	pjp->up_neigh_addr = n->address;
	pjp->reserved = 0;
	bcopy(jpholdtime, &pjp->holdtime, 2);
	pjp->reserved2 = 0;
	pjp->numgroups = 1;
	pjp->addrlen = ADDR_LEN;
	pjp->maddrlen = 4;
	ma = pjp->list;
	ma->gaddr = mrtPtr->group->group;
	ma->gaddrmask = ~0;
	ma->njoins = 0;
	ma->nprunes = 1;
	cp = ma->eaddrs;

	cp[0] = 1;		/* Sparse */
	cp[1] = 0x40 | 32;	/* RPT and addrlen */
	memcpy(cp, &mrtPtr->source->source, 4);

	send_pim_v1(pimv1_send_buf, v->prefsrc, allrouters_group,
		    PIMv1_JOIN_PRUNE, sizeof(*pjp) + sizeof(*ma) + 6,
		    v->ifindex, 0, -1);
	return (TRUE);
}

void send_null_register_v1(struct mrtentry *r)
{
	register struct ip *ip;
	char *buf;
	int pktlen = 0;
	__u32 dest;

	buf = pimv1_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN;

	ip = (struct ip *) (buf);
	memset(ip, 0, sizeof(*ip));
	ip->ip_src.s_addr = r->source->source;	/* set src/dst in dummy hdr */
	ip->ip_dst.s_addr = r->group->group;

	pktlen += MIN_IP_HEADER_LEN;	/* include the dummy ip header */

	dest = r->group->active_rp->rpaddr;

	if (debug > 1)
		log(LOG_DEBUG, 0, "send_null_register: (%s,%s) to RP:%s",
		    inet_fmt(r->source->source, s1),
		    inet_fmt(r->group->group, s2),
		    inet_fmt(dest, s3));

	send_pim_v1(pimv1_send_buf,
		    0,
		    dest,
		    PIMv1_REGISTER,
		    pktlen,
		    0, -1, MAXTTL);
}


int accept_assert_v1(__u32 src, __u32 dst, char *buf, int datalen)
{
	vifi_t vifi;
	struct nbrentry *n;
	struct pim_assert *pim = (void*)buf;

	if ((vifi = vifi_by_index(igmp_ifindex)) == MAXVIFS)
		return FALSE;
	n = neigh_lookup(vifi, src, 1);
	if (n == NULL)
		return FALSE;

	return accept_assert_gen(vifi, n, pim->saddr, pim->gaddr,
				 ntohl(pim->metricpref), ntohl(pim->metric));
}


/* static : hack. look at mrt.c */
void send_assert_v1(struct mrtentry *mrtPtr, struct vifentry *v)
{
	struct pim_assert * pa = (void*)(pimv1_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN);

	pa->gaddr = mrtPtr->group->group;
	pa->gmask = ~0;
	pa->saddr = mrtPtr->source->source;
	pa->metric = htonl(mrtPtr->source->metric);
	pa->metricpref = htonl(mrtPtr->source->preference);
	if (mrtPtr->flags & MRTF_RP)
		pa->metricpref |= __constant_htonl(0x80000000);
	send_pim_v1(pimv1_send_buf, v->prefsrc, allrouters_group,
		    PIMv1_ASSERT, sizeof(*pa), v->ifindex, FALSE, 1);
}

void accept_join_prune_v1(__u32 src, __u32 dst, char *p, int datalen)
{
	u_short flags;
	u_long source, group;
	vifi_t vifi;
	struct vifentry *v;
	__u32 target;
	int len, num_joins, num_prunes;
	int i, j;
	u_char addrlen, num_grps;
	struct mga_join_prune *jplist;
	struct nbrentry *n;
	struct pim_jp *pim1 = (struct pim_jp *) p;

	if ((vifi = vifi_by_index(igmp_ifindex)) == MAXVIFS ||
	    (n = neigh_lookup(vifi, src, 1)) == NULL) {
		log(LOG_DEBUG, 0, "Not directly cnctd, ignoring (%s)!\n", inet_fmt(src, s1));
		return;		/* if not directly cnctd */
	}
	if (debug > 3) {
		log(LOG_DEBUG, 0, "accept_join_prune_v1: src %s, dst %s, on vifi %d, upstream nbr: %s",
		    inet_fmt(src, s1),
		    inet_fmt(dst, s2),
		    vifi,
		    inet_fmt(pim1->up_neigh_addr, s1));
	}
	addrlen = pim1->addrlen;

	v = &viflist[vifi];

	target = pim1->up_neigh_addr;

	if (!is_local_on(target, v)) {
		if (debug > 1)
			log(LOG_DEBUG, 0, "join_prune_v1: upstream nbr: %s on %s",
			    inet_fmt(target, s1),
			    v->name);
		return;
	}

	/* I am the target of this join, so process the message */
	num_grps = pim1->numgroups;
	log(LOG_DEBUG, 0, "Processing %d groups\n", num_grps);
	/* copy the holdtime in the join, into the oif timeout of the
	 * corresponding vif
	 */
	oifTimeout[vifi] = htons(pim1->holdtime);
	p += (char*)&pim1->list - (char*)pim1;
	len = 0;
	jplist = (struct mga_join_prune *) p;


	for (i = 0; i < num_grps; i++) {
		if (debug > 3)
			log(LOG_DEBUG, 0, "Num of Groups = %u", num_grps);

		group = jplist->gaddr;
		p += 8;
		num_joins = ntohs(jplist->njoins);
		num_prunes = ntohs(jplist->nprunes);
		log(LOG_DEBUG, 0, "%d-Joins, %d-Prunes", num_joins, num_prunes);
		p += 4;

		for (j = 0; j < num_joins; j++) {
			/* this is assuming the reserved is 0 */

			flags = (((((p[1]) & (1 << 7)) ? MRTF_WC : 0) |
				  ((p[1] & (1 << 6)) ? MRTF_RP : 0))
				 & (0x00ff));
			p += 2;
			memcpy(&source, p, addrlen);
			p += addrlen;

			do_accept_join(vifi, source, group, flags);
		}

		for (j = 0; j < num_prunes; j++) {

			if (debug > 3) {
				log(LOG_DEBUG, 0, "Num of Prunes = %u", num_prunes);
				if (num_joins > 5)
					return;
			}
			flags = (((((p[1]) & (1 << 7)) ? MRTF_WC : 0) |
				  ((p[1] & (1 << 6)) ? MRTF_RP : 0))
				 & (0x00ff));
			p += 2;
			memcpy(&source, p, addrlen);
			p += addrlen;

			do_accept_prune(vifi, source, group, flags);
		}
	}
}


int event_join_prune_v1(struct mrtentry *r, char join,
			struct nbrentry *n, struct grpentry *g,
			struct vifentry *v)
{
	char *buf, *bufPtr;
	int datalen;
	u_long num_joins, num_prunes;
	struct pim_jp *pim1;
	char join_buf[ADDRSIZE * MAXSrc], prune_buf[ADDRSIZE * MAXSrc];
	char *jptr, *pptr;
	struct encoded_addr *eaddr;
	u_char num_grps;

	buf = pimv1_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN;
	pim1 = (struct pim_jp *) buf;
	pim1->up_neigh_addr = n->address;
	pim1->reserved = 0;
	pim1->holdtime = htons(OUTGOING_TIMEOUT);
	pim1->reserved2 = 0;
	pim1->maddrlen = 4;	/* bytes */
	pim1->addrlen = ADDR_LEN;

	jptr = join_buf;
	pptr = prune_buf;
	datalen = 12;		/* Three quads just filled */
	num_joins = 0;
	num_prunes = 0;
	num_grps = 0;
	if ((r->flags & MRTF_RP) && !(r->flags & MRTF_WC) && (!join))
		/* RPbit Entry */
		encode_source_v1(&pptr, r, TRUE, SRC_MSKLEN);
	else {
		if ((((r->outgoing | g->leaves) == EMPTY_VIF) && (!join))) {
			num_prunes++;
			encode_source_v1(&pptr, r, FALSE, SRC_MSKLEN);
			if (debug > 4)
				log(LOG_DEBUG, 0, "event_join_prune_v1:pruning Src %s, Grp %s, entry flags %x",
				    inet_fmt(r->source->source, s1),
				    inet_fmt(r->group->group, s2),
				    r->flags);
		} else {	/* Right now it doesn't take care of Jbit */
			num_joins++;

			encode_source_v1(&jptr, r, FALSE, SRC_MSKLEN);

			if (debug > 4)
				log(LOG_DEBUG, 0, "event_join_prune_v1:joining Src %s, Grp %s, entry flags 0x%x",
				    inet_fmt(r->source->source, s1),
				    inet_fmt(r->group->group, s2),
				    r->flags);
		}
	}

	if (debug > 3)
		log(LOG_DEBUG, 0, "Num Joins %u, Num Prunes %u",
		    num_joins, num_prunes);

	/* if the total number is zero, don't send */
	if ((num_joins + num_prunes) == 0)
		return (FALSE);
	num_grps++;
	/* Pointer to mga_join_prune struct */
	bufPtr = &buf[datalen];
	pim1->list[0].gaddr = r->group->group;
	pim1->list[0].gaddrmask = (unsigned long) 0xffffffff;
	datalen += 8;
	/* Copy the LS two bytes */
	pim1->list[0].njoins = htons(num_joins);
	pim1->list[1].nprunes = htons(num_prunes);
	datalen += 4;
	eaddr = (struct encoded_addr *) join_buf;
	log(LOG_DEBUG, 0, "Datalen=%d , num_joins=%d", datalen, num_joins);

	bcopy(join_buf, buf + datalen, num_joins * ADDRSIZE);
	bcopy(prune_buf, buf + datalen + num_joins * ADDRSIZE, num_prunes * ADDRSIZE);

	datalen += (num_joins + num_prunes) * ADDRSIZE;

	pim1->numgroups = num_grps;

	if (debug > 4)
		log(LOG_DEBUG, 0,
		    "event_join_prune_v1:TX triggered J/P to all pim routers");

	send_pim_v1(pimv1_send_buf, v->prefsrc, allrouters_group, PIMv1_JOIN_PRUNE,
		    datalen, v->ifindex, FALSE, 1);

	log(LOG_DEBUG, 0, "CHECKING: NGroups=%d First=%s:%s\n",
	    pim1->numgroups,
	    inet_fmt(pim1->list[0].gaddr, s1),
	    inet_fmt(pim1->list[0].gaddrmask, s2));
	return (TRUE);
}

/* function name: periodic_join_prune_v1
   input: a neighbor entry pointer, and a virtual interface
   output: int (the number of groups for which a Join/Prune
   message has been sent)
   operation: Sends the periodic join/prune message to the
   neighbor n, and returns number of groups for which
   message has been sent.
   comments: called by periodic_vif_join_prune
   calls join_or_prune
 */
int periodic_join_prune_v1(struct nbrentry *n, vifi_t vifi)
{
	u_long num_joins, num_prunes;
	int datalen;
	u_long group;
	register struct mrtentry *r, *r2;
	struct grpentry *g;
	struct pim_jp *pim1;
	struct mga_join_prune *join_prune1;
	char *jptr, *pptr;
	char join_buf[ADDRSIZE * MAXSrc], prune_buf[ADDRSIZE * MAXSrc];
	char *buf, *bufPtr;
	u_char num_grps;
	int j_or_p;

	num_grps = 0;
	buf = pimv1_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN;
	pim1 = (struct pim_jp *) buf;
	if (n != NULL) {	/* The upstream neighbor Address */
		pim1->up_neigh_addr = n->address;
	} else {
		tracef(TRF_PIM1, "PIMv1: periodic_join_prune_v1 on %s: No PIM upstream neighbor",
		       viflist[vifi].name);
		/* if the upstream neighbor is not a PIM neighbor, don't
		   send any messages... return an error code (-1) */
		return (-1);
	}
	pim1->reserved = 0;
	pim1->holdtime = htons(OUTGOING_TIMEOUT);
	pim1->reserved2 = 0;
	pim1->maddrlen = 4;	/* bytes */
	pim1->addrlen = ADDR_LEN;
	datalen = 12;

	forall_groups(g) {
		num_joins = 0;
		num_prunes = 0;
		group = g->group;
		r = g->mrtlink;
		jptr = join_buf;
		pptr = prune_buf;
		for (r2 = r; r2 != NULL; r2 = r2->grpnext) {
			/* Check if its time to send the periodic join_prune */

			/* may want to maintain statistics about sending joins,
			   for debugging, testing and performance analysis */
			if (debug > 10)
				log(LOG_DEBUG, 0, "Join timer on %s = %u",
				    inet_fmt(group, s1),
				    r2->join_prune_timer);

			if ((r2->join_prune_timer += TIMER_INTERVAL)
			    < 15 /* JOIN_PRUNE_INTERVAL */ )
				continue;

			r2->join_prune_timer = 0;	/* Reset the timer */

			/* one of my address */
			if (local_address(r2->source->source) != MAXVIFS)
				continue;

			j_or_p = join_or_prune(n, vifi, r2);
			if (debug > 10)
				log(LOG_DEBUG, 0, "Periodic J/P: %s",
				    j_or_p == NONE ? "NONE" :
				    (j_or_p == PRUNE ? "PRUNE" : "JOIN"));

			if (j_or_p == NONE)
				continue;
			if (j_or_p == PRUNE) {
				tracef(TRF_PIM1, "\nsending Prunes for (%s,%s)",
				       inet_fmt(r2->source->source, s1),
				       inet_fmt(r2->group->group, s2));


				num_prunes++;
				encode_source_v1(&pptr, r2, TRUE, SRC_MSKLEN);
			} else {
				/* it's a Join */
				tracef(TRF_PIM1, "\nsending Join for (%s,%s)",
				       inet_fmt(r2->source->source, s1),
				       inet_fmt(r2->group->group, s2));

				num_joins++;

				encode_source_v1(&jptr, r2, FALSE, SRC_MSKLEN);
			}
		}
		if ((num_joins + num_prunes) == 0)
			continue;
		num_grps++;

		bufPtr = (char *) &buf[datalen];
		join_prune1 = (struct mga_join_prune *) bufPtr;
		join_prune1->gaddr = group;
		join_prune1->gaddrmask = (unsigned long) 0xffffffff;
		datalen += 8;

		join_prune1->njoins = htons(num_joins);
		join_prune1->nprunes = htons(num_prunes);
		datalen += 4;

		bcopy(join_buf, buf + datalen, num_joins * ADDRSIZE);
		bcopy(prune_buf, buf + datalen + num_joins * ADDRSIZE, num_prunes * ADDRSIZE);
		datalen += (num_joins + num_prunes) * ADDRSIZE;
		continue;
	} endforall_groups(g);

	if (num_grps == 0)
		return num_grps;
	pim1->numgroups = num_grps;

	tracef(TRF_PIM1, "PIMv1: periodic_join_prune_v1 on %s: to upstream neighbor %s",
	       viflist[vifi].name, inet_fmt(n->address, s1));
	send_pim_v1(pimv1_send_buf, viflist[vifi].prefsrc, allrouters_group,
		    PIMv1_JOIN_PRUNE, datalen, viflist[vifi].ifindex, FALSE, 1);
	return num_grps;
}


void send_register_v1(__u32 dest, struct ip *ip, int pktlen)
{
	char *buf;

	buf = pimv1_send_buf + MIN_IP_HEADER_LEN + IGMP_MINLEN;

	/* Copy the data packet at the back of the register packet */
	pktlen += ntohs(ip->ip_len);
	bcopy((char *) ip, buf, pktlen);

	send_pim_v1(pimv1_send_buf,
		    0,
		    dest,
		    PIMv1_REGISTER,
		    pktlen,
		    0, -1, MAXTTL);
}
