/* Copyright (c) 1996 University of Southern California
 * All rights reserved.
 *
 * Do NOT redistribute.
 *
 */
/*
 * Procedures for the Daemon part of PIM-SMv2.0,
 * Protocol Independent Multicast, Sparse-Mode version 2.0
 *
 * MROUTING 2.0
 */
#include <sys/param.h>
#include <sys/file.h>
#include <sys/mbuf.h>
#include <net/if_dl.h>
#include <net/route.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdlib.h>
#include <strings.h>
#include <paths.h>
#include "defs.h"

union sockunion {
	struct sockaddr sa;
	struct sockaddr_in sin;
	struct sockaddr_dl sdl;
} so_dst, so_gate, so_mask, so_genmask, so_ifa, so_ifp;
typedef union sockunion *sup;
int s;
int rtm_addrs, pid;
struct rt_metrics rt_metrics;
u_long rtm_inits;

struct rpfctl *getmsg(register struct rt_msghdr *, int);

#define ROUNDUP(a) ((a) > 0 ? (1 + (((a) - 1) | (sizeof(__uint64_t) - 1))) \
		    : sizeof(__uint64_t))

#define ADVANCE(x, n) (x += ROUNDUP(_FAKE_SA_LEN_DST(n)))

/* open and initialize the routing socket */
int init_routesock()
{
	pid = getpid();
	s = socket(PF_ROUTE, SOCK_RAW, 0);
	if (s < 0) {
		log(LOG_ERR, 0, "\nRouting socket error");
		return -1;
	}
	return 0;
}

struct {
	struct rt_msghdr m_rtm;
	char m_space[512];
} m_rtmsg;

/* get the rpf neighbor info */
struct rpfctl *
 k_req_incoming(u_long source)
{
	int flags = RTF_STATIC;
	register sup su;
	u_long val;
	static int seq;
	int rlen;
	register char *cp = m_rtmsg.m_space;
	register int l;
	struct rpfctl *rpfptr, rpfinfo;

#define NEXTADDR(w, u) \
        if (rtm_addrs & (w)) {\
            l = _FAKE_SA_LEN_DST(&u.sa); l = ROUNDUP(l);\
            bcopy((char *)&(u), cp, l); cp += l;\
        }

	/* initialize */
	rpfinfo.rpfneighbor.s_addr = INADDR_ANY;

	/* check if local address before calling the routing socket */
	if ((rpfinfo.iif = local_address(source)) != MAXVIFS) {
		rpfinfo.rpfneighbor.s_addr = source;
		return (&rpfinfo);
	}
	/* check if directly connected */
	if ((rpfinfo.iif = find_vif(source)) != MAXVIFS) {
		rpfinfo.rpfneighbor.s_addr = source;
		return (&rpfinfo);
	}
	/* prepare the routing socket params */
	rtm_addrs |= RTA_DST;
	rtm_addrs |= RTA_IFP;
	su = &so_dst;
	su->sa.sa_family = AF_INET;
	su->sin.sin_family = AF_INET;
	su->sin.sin_addr.s_addr = source;
	if (inet_lnaof(su->sin.sin_addr) == INADDR_ANY) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "k_req_incoming: Invalid source %s",
			    inet_fmt(source, s1));
		return (&rpfinfo);
	}
	so_ifp.sa.sa_family = AF_LINK;
	flags |= RTF_UP;
	flags |= RTF_HOST;
	flags |= RTF_GATEWAY;
	errno = 0;
	bzero((char *) &m_rtmsg, sizeof(m_rtmsg));

#define rtm m_rtmsg.m_rtm
	rtm.rtm_type = RTM_GET;
	rtm.rtm_flags = flags;
	rtm.rtm_version = RTM_VERSION;
	rtm.rtm_seq = ++seq;
	rtm.rtm_addrs = rtm_addrs;
	rtm.rtm_rmx = rt_metrics;
	rtm.rtm_inits = rtm_inits;

	NEXTADDR(RTA_DST, so_dst);
	NEXTADDR(RTA_GATEWAY, so_gate);
	NEXTADDR(RTA_NETMASK, so_mask);
	NEXTADDR(RTA_GENMASK, so_genmask);
	NEXTADDR(RTA_IFP, so_ifp);
	NEXTADDR(RTA_IFA, so_ifa);
	rtm.rtm_msglen = l = cp - (char *) &m_rtmsg;

	if ((rlen = write(s, (char *) &m_rtmsg, l)) < 0) {
		if (debug > 4) {
			if (errno == ESRCH)
				log(LOG_DEBUG, 0, "Writing to routing socket: no such route\n");
			else
				log(LOG_DEBUG, 0, "Error writing to routing socket");
		}
		/* we failed so return a struct with error values */
		return (&rpfinfo);
	}
	do {
		l = read(s, (char *) &m_rtmsg, sizeof(m_rtmsg));
	} while (l > 0 && (rtm.rtm_seq != seq || rtm.rtm_pid != pid));

	if (l < 0) {
		if (debug > 4)
			log(LOG_DEBUG, 0, "Read from routing socket failed");
		/* we failed so return a struct with error values */
		return (&rpfinfo);
	}
	if (rpfptr = getmsg(&rtm, l)) {
		rpfinfo.rpfneighbor.s_addr = rpfptr->rpfneighbor.s_addr;
		rpfinfo.iif = rpfptr->iif;
	}
#undef rtm
	return (&rpfinfo);
}

struct rpfctl *
 getmsg(register struct rt_msghdr *rtm, int msglen)
{
	struct sockaddr *dst = NULL, *gate = NULL, *mask = NULL;
	struct sockaddr_dl *ifp = NULL;
	register struct sockaddr *sa;
	register char *cp;
	register int i;
	struct in_addr in;
	struct rpfctl rpfinfo;
	vifi_t vifi;
	struct vifentry *v;

	in = ((struct sockaddr_in *) &so_dst)->sin_addr;
	if (debug > 5)
		log(LOG_DEBUG, 0, "route to: %x", in.s_addr);
	cp = ((char *) (rtm + 1));
	if (rtm->rtm_addrs)
		for (i = 1; i; i <<= 1)
			if (i & rtm->rtm_addrs) {
				sa = (struct sockaddr *) cp;
				switch (i) {
				case RTA_DST:
					dst = sa;
					break;
				case RTA_GATEWAY:
					gate = sa;
					break;
				case RTA_NETMASK:
					mask = sa;
					break;
				case RTA_IFP:
					if (sa->sa_family == AF_LINK &&
					    ((struct sockaddr_dl *) sa)->sdl_nlen)
						ifp = (struct sockaddr_dl *) sa;
					break;
				}
				ADVANCE(cp, sa);
			}
	if (dst) {
		in = ((struct sockaddr_in *) dst)->sin_addr;
		if (debug > 5)
			log(LOG_DEBUG, 0, " destination is: %x", in.s_addr);
	}
	if (gate && rtm->rtm_flags & RTF_GATEWAY) {
		in = ((struct sockaddr_in *) gate)->sin_addr;
		if (debug > 5)
			log(LOG_DEBUG, 0, " gateway is: %x", in.s_addr);
		rpfinfo.rpfneighbor = in;
	}
	for (vifi = 0, v = viflist; vifi < numvifs; ++vifi, ++v) {
		/* get the number of the interface by matching the name */
		if (!(strncmp(v->name, ifp->sdl_data, ifp->sdl_nlen)))
			break;
	}

	if (debug > 5)
		log(LOG_DEBUG, 0, " iif is %d", vifi);

	rpfinfo.iif = vifi;

	if (vifi >= numvifs)
		return NULL;	/* invalid iif */

	return (&rpfinfo);
}
